#!/usr/bin/env python

# 
#  Copyright (C) 2019  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import sys
import select
import numpy as np
import matplotlib.pyplot as plt
import ast
from pycrates import *
import matplotlib

import time
from matplotlib import _pylab_helpers
from matplotlib.rcsetup import interactive_bk as _interactive_bk

matplotlib.interactive(True)


def add_curve(**kwargs):
    """ add a curve of the specified data (using dm syntax through crates)

        kwargs must contain 
        - filespec: path qualified table file name including dm syntax '[cols col1, col2, col3]'
            where col1=x data, col2=y data, optional cols=errs. If no column specification
            is provided the first 3 columns are used. 
        - behavior: behavior of subsequent plot requests
            0 = create a new window for the plot
            1 = overplot (add curve to current existing plot)

        kwargs may optionally contain
        - dictionary of matplotlib properties
    """
    if kwargs.__contains__('behavior'):
        overplot = int(kwargs['behavior'])
        kwargs.pop('behavior')
    else:
        overplot = 1
    if kwargs.__contains__('filespec'):
        fname = kwargs['filespec'] 
        kwargs.pop('filespec')
    else:
        return
     
    if len(fname) == 0:
        return
    try:
        ff = TABLECrate(fname, mode='r')
    except:
        raise IOError("unable to open file '%s'" % fname)
    if ff == None:
        return
    nc = ff.get_ncols()
    c2 = None 
    if nc == 0:
        raise IOError("No columns to plot in %s" % fname)
    if nc > 3:
        raise IOError("Too many columns to plot in %s" % fname)
    if nc == 1:
        tmpcol = ff.get_column(0)
        if tmpcol.is_vector():
            vnames = tmpcol.get_cptslist()
            if len(vnames) == 2:
                c0 = tmpcol.get_vector_component(vnames[0])
                c1 = tmpcol.get_vector_component(vnames[1])
            elif len(vnames) == 3:
                c0 = tmpcol.get_vector_component(vnames[0])
                c1 = tmpcol.get_vector_component(vnames[1])
                c2 = tmpcol.get_vector_component(vnames[2])
            else:
                raise IOError("too many components in the vector column to plot in %s" % fname)
        else:
            raise IOError("insufficient number of columns to plot in %s" % fname)
    else:
        tmpcol = ff.get_column(0)
        if tmpcol.is_vector():
            vnames = tmpcol.get_cptslist()
            if len(vnames) > 2:
                raise IOError("too many columns or vector components to plot in %s" % fname)
            else:
                c0 = tmpcol.get_vector_component(vnames[0])
                c1 = tmpcol.get_vector_component(vnames[1])
            c2 = ff.get_column(1)
            if c2.is_vector():
                if c2.get_size() > 1:
                    raise IOError("too many columns or vector components to plot in %s" % fname)
                else:
                   c2 = c2.get_column(0)
        else:
            c0 = tmpcol
            c1 = ff.get_column(1)
            if c1.is_vector():
                if c1.get_size() == 2: 
                    if nc > 2: 
                        raise IOError("too many columns or vector components to plot in %s" % fname)
                    c2 = c1.get_vector_component(1)
                    c1 = c1.get_vector_component(0)
                elif c1.get_size() == 1:
                    c1 = c1.get_get_vector_component(0)
                    if nc == 3:
                        c2 = ff.get_column(2)
                else:
                    raise IOError("Incorrect number of columns or vector components to plot in %s" % fname)
            if nc > 2:
                c2 = ff.get_column(2)
                if c2.get_size() > 1: 
                    raise IOError("too many columns or vector components to plot in %s" % fname)

    xx = c0.values
    yy = c1.values
    xlab = c0.name
    ylab = c1.name
    cax = None
    if overplot:
        aa = plt._pylab_helpers.Gcf.get_active()
        cax = plt.gca();
        fig = plt.gcf()   
        if overplot == 2 and cax:
            # clear existing plot and reset labels 
            plt.cla()
            cax.set_xlabel(xlab)
            cax.set_ylabel(ylab)
    if cax == None:
        fig,cax = plt.subplots()
        cax.set_xlabel(xlab)
        cax.set_ylabel(ylab)
    if c2:
        zz = c2.values
        ll = cax.errorbar(xx, yy, zz, **kwargs)
    else:
        ll = cax.plot(xx, yy, **kwargs)
    if fig:
        fig.canvas.draw()
        fig.canvas.flush_events()
    plt.show(block=False)


def add_histogram(**kwargs):
    """ add histogram of the specified data (using dm syntax through crates)

        kwargs must contain 
        - filespec: path qualified table file name including dm syntax '[cols col1, col2, col3]'
            where col1=x data, col2=y data, optional cols=errs. If no column specification
            is provided the first 3 columns are used. 
        - behavior: behavior of subsequent plot requests
            0 = create a new window for the plot
            1 = overplot (add curve to current existing plot)

        kwargs may optionally contain
        - dictionary of matplotlib properties
    """
    if kwargs.__contains__('behavior'):
        overplot = int(kwargs['behavior'])
        kwargs.pop('behavior')
    else:
        overplot = 1
    if kwargs.__contains__('filespec'):
        fname = kwargs['filespec'] 
        kwargs.pop('filespec')
    else:
        return 

    if len(fname) == 0:
        return
    try:
        ff = TABLECrate(fname, mode='r')
    except:
        raise IOErrror("unable to open file '%s'" % fname)
    if ff != None:
        nc = ff.get_ncols()
        if nc < 1:
            raise IOErrror("insufficient number of columns to plot in %s" % fname)
        else:
            c0 = ff.get_column(0)

        xx = c0.values
        xlab = c0.name
        cax = None
        if overplot:
            aa = plt._pylab_helpers.Gcf.get_active()
            cax = plt.gca();
            fig = plt.gcf()
            if overplot == 2 and cax:
                # clear existing plot and reset labels 
                plt.cla()
        if cax == None:
            fig,cax = plt.subplots()
        if kwargs.get("label") == None:
            kwargs.setdefault("label", xlab)
        ll = cax.hist(xx, **kwargs)
        cax.legend()
        if fig:
            fig.canvas.draw()
            fig.canvas.flush_events()
        plt.show(block=False)


def my_pause(interval, focus_figure=True):
    """
       replacement for plt.pause since plt.pause forces a figure focus 
       which will force the desktop workspace to be the one the plot resides
       in (unexpected if the user changes explicitly switches workspaces)
    """
    backend = matplotlib.rcParams['backend']
    if backend in _interactive_bk:
        figManager = _pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            if focus_figure:
                plt.show(block=False)
            try:
                canvas.start_event_loop(interval)
            except:
                return
            return
    # No on-screen figure is active, so sleep() is all we need.
    time.sleep(interval)

#------------------------------------
# main routine
#        
#-----------------------------------
if __name__ == "__main__":
    done = False;
    while(done !=True):
        read_list = [sys.stdin]
        while read_list:
            ready = select.select(read_list, [], [], 0.1)[0]
            if not ready:
                if plt.get_fignums():
                    plt.draw()
                    my_pause(5, False);
            else:
                for file in ready:
                    line = file.readline()
                    if not line: # EOF, remove file from input list
                        read_list.remove(file)
                    rr = line.strip()
                    if rr.lower() == "quit":
                        done = True
                    else:
                        # need to pop off 'viztype' to determine if
                        # plotting curve or histogram 
                        myargs = eval(rr)
                        if myargs.__contains__('viztype'):
                            plottype = myargs['viztype']
                            myargs.pop('viztype')
                        if plottype == "curve":
                            add_curve(**myargs)
                        elif plottype == "histogram":
                            add_histogram(**myargs)

