# 
#  Copyright (C) 2019  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#



""" 
test pypixlib modules
"""

from __future__ import print_function

from pydoc import help   #for run-time help
import unittest
import sys,os
import numpy as np 
from pixlib import *

def show_whoami():
    print('\n---', sys._getframe(1).f_code.co_name)

def usage():
    """Show the usage with example for Caldb python module."""
    
    print("""\nUsage:\n\
           % python test_pixlib.py [opt] [-v]
           opt:
              -aim:   test aimpoint
              -det:   test detector
              -def:   test default
              -gra:   test grating
              -cor:   test coordinates transfer
              -gen:   test general funtions
              -ext:   test extened chip transfers
              -fai:   test expected failures
                -u:   this message
                -v:   debug option\n""")



DEBUG=0
class TestPixlib(unittest.TestCase):
    def setUp(self):
        global DEBUG
        if "-v" in  sys.argv:
            DEBUG=1
            #sys.argv.remove("-v")
            
    def tearDown(self):
        """ un-import caldb4 ?"""
        nm=['pix', '']
        while True:
            if nm[0] in sys.modules:
                del (sys.modules[nm[0]])
                del (nm[0])
            else:
                break
        return
   

#=======================================
# aimpoint - test aimpoint
#
# AI1    : (   -0.782     0.000  -237.500 )
# HI1    : (   -1.040     0.000   126.985 )
# HS1    : (   -1.430     0.000   250.456 )
# HSI1   : (   -1.430     0.000   250.456 )
# A2C0   : (   -1.430     0.000   250.456 )
# AS1    : (   -0.684     0.000  -190.133 )
# AI2    : (   -0.782     0.000  -233.592 )
# HS2    : (   -1.533     0.000   250.456 )

#========================================
    def test_aimpoint(self):
        p = Pixlib("chandra")
        print("\nTest pixlib Aimpoints ","-"*10)


        print("-- Default aimpoint: ")
        print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord) )

        print("\n-- Set aimpoint in alias:\n")
        p.aimpoint="AI2"
        print(repr(str(p.aimpoint))," = ",pixlib.set_decim_pts(p.aimcoord) )
    
        p.aimpoint="hi1"
        print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord) )

        p.aimpoint="hs1"
        print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord) )

        p.aimpoint="as1"
        print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord) )
  
        p.aimpoint="hs2"
        print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord) )

        print("\n-- Set aimpoint by value:")
        p.aimpoint = -0.78234819833843305,0.0,-233.5924630914318
        aimname=p.aimpoint
        if p.isokay == 1:
            print(repr(str(aimname)), " = ",pixlib.set_decim_pts(p.aimcoord), "\n")
        else:
            print(p)
     

        print("\n-- Expected error returned from setting 'hsi1, a2c0, mypnt' -> ")
        p.aimpoint="HSI1"
        if p.isokay == 1:
            print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord))
        else:
            print("\t"+p.errmsg)

        p.aimpoint="A2C0"
        if p.isokay == 1:
            print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord))
        else:
            print("\t"+p.errmsg)

        p.aimpoint="mypnt"
        if p.isokay == 1:
            print(repr(str(p.aimpoint)), " = ",pixlib.set_decim_pts(p.aimcoord) )
        else:
            print("\t"+p.errmsg+"\n")

        p.close
 
 #Note:
 # there may be 'Segmentation fault' at the exit of the routine
 # it is probably due to some of compiling options
 #

 
    def test_detector(self):
        p = Pixlib("chandra")
        print("\nTest pixlib Detectors ","-"*10)
        detid=p.detector
        print("-- Default detector: ")
        print(repr(str(detid)), " -> expected: acis")

        print("\n-- Set detector in string:")
        p.detector="acis-i"
        detid=p.detector
        print(repr(str(detid)), "    -> expected: acis")

        p.detector="acis-s1"
        detid=p.detector
        print(repr(str(detid)), "    -> expected: acis")

        p.detector="hrc-i"
        detid=p.detector
        print(repr(str(detid)), "   -> expected: hrc-i")

        p.detector="hrc-s"
        detid=p.detector
        print(repr(str(detid)), "   -> expected: hrc-s")

        p.detector="hsi-i"
        detid=p.detector
        print(repr(str(detid)), "     -> expected: hsi")

        p.detector="acis-2c"
        detid=p.detector
        print(repr(str(detid))," -> expected: acis-2c")

        p.close
  
    def test_default(self):
        p = Pixlib()
        print("\nTest pixlib defaults ","-"*10)

        print("detector: ", repr(str(p.detector)), "             -> expected: 'acis'")
        print("aimpoint: ", repr(str(p.aimpoint)), "              -> expected: 'ai1'")
        print(" tdetsys: ", repr(str(p.tdetsys)), "         -> expected: 'ACIS-2.2'")
        print("   fpsys: ", repr(str(p.fpsys)), "           -> expected: 'FP-1.1'")
        print("  gdpsys: ", repr(str(p.gdpsys)), "      -> expected: 'ASC-GDP-1.1'")
        print(" grating: ", repr(str(p.grating)), "              -> expected: 'HEG'")
  
        p.close
    
    def test_grating(self):
        p = Pixlib()
        print("\n-- Test pixlib Grating ","-"*10)

        print("\n-- See what grating defaults are ...")
        print("Grating  = ", p.grating)
        if p.grt_zo[0]==None:
            print("grt_ZO   = ", p.grt_zo) 
        else:
            print("grt_ZO   = ", pixlib.set_decim_pts(p.grt_zo) )
        print("grt_prop = ", pixlib.set_decim_pts(p.grt_prop, 4))
   
        print("\n-- Set gratings arm ...")
        p.grating='leg'
        print("Grating = ", p.grating)
        print("grt_prop = ", pixlib.set_decim_pts(p.grt_prop, 4))

        print("\n-- Set gratings arm and order ...")
        p.grating=['meg',2]
        print("Grating   = ", p.grating)
        print("grt_order = ", p.grt_order)
        print("grt_prop  = ", pixlib.set_decim_pts(p.grt_prop, 4))
 
        print("\n-- GDP sys ...")
        print("default GDP sys = ", p.gdpsys)
        p.gdpsys = "gdp-1.2"
        print("GDP sys = ", p.gdpsys)
        p.gdpsys = "gdp-1.3"
        print("GDP sys = ", p.gdpsys)
        p.gdpsys = "gdp-1.4"
        print("\n-- Expected error from setting gdp-1.4 ->")
    #ERROR: Unsupported grating diffraction plane system name, at 'gdp-1.4'.

        if p.isokay != 1:
            print(p)
        else:
            print("GDP sys = ", p.gdpsys)
    
        print("-- Expected error from getting energy without argument ->")
        grt_energy = p.grt_energy()
        if p.isokay != 1:
            print(p)
        else:
            print("grt-energy = ", pixlib.set_decim_pts(grt_energy))
       
        grt_energy = p.grt_energy((0.3,0.5))
        print("grt-energy = ",  pixlib.set_decim_pts(grt_energy) )
        grt_wavelen =p.grt_wavelen()
        if p.isokay != 1:
            print(p)
        else:
            print("grt-wavelen = ",  pixlib.set_decim_pts(grt_wavelen))

        print("grt-energy = ",  pixlib.set_decim_pts(p.grt_energy()))
        print("\nAll curr inputs/updates:", p.curstat())
    
        p.close
  
    def test_general(self):
        p = Pixlib("chandra")
        print("\n-- Test pixlib general functions ","-"*10)

        print("\nfoc length: ",pixlib.set_decim_pts(p.flength,1), \
            "\t -> expected: 10070.0")
        stg=1.1,2.2,3.3
        ang=23.0
        p.mirror=(stg,ang)
        print("\nstg pos&angle: (",pixlib.set_decim_pts(stg,1),",", pixlib.set_decim_pts(ang,1),")" \
            "\t -> expected: ( (1.1, 2.2, 3.3) , 23.0 )")

        p.grt_zo=3.0,2.0
        print("\nzero-order in error: ",p.grt_zo)
        print(p)
        p.grt_zo=3.0,2.0,1.0
        print("zero-order: ",p.grt_zo, \
            "\t -> expected:  (3.0, 2.0, 1.0)")

        p.grt_order='5'
        print("grt-order in error: ",p.grt_order)
        print(p)
        p.grt_order=5
        print("grt-order: ",p.grt_order, \
            "\t -> expected:  5")
    
        p.grating='leg',4
        print("grating and order: ",p.grating, p.grt_order,\
            "\t -> expected:  leg 4")

        print("\nfp-scale: ", pixlib.set_decim_pts(p.fp_scale,3),\
            "\t -> expected:  0.492\n")
    
        p.close

    
    
    def test_transfer(self):
        p = Pixlib("chandra")
        print("\nTest pixlib coords conversion ","-"*10)


    #ACIS-2.2  ACIS-I3 ID  3  CHIP: (783.8,894.0) TDET: (  5132.00,   3844.80)
    #ACIS-2.2        ACIS-S2 ID  6    DET_coords: (  3658.80,   1702.00)

        id=2 
        chip=140.47,58.37 
        p.aimpoint="AI2"
        print("set aimpoint at ",p.aimpoint)
        lsi=p.chip2lsi([id, chip])
    #lsi= (1.2287446656894812, -25.155139506394796, -5.3452310290900176)
        print("chip2lsi:\n\tid,chip=", id,chip,"\n\tlsi=",pixlib.set_decim_pts(lsi))

        p.fpsys="FP-1.0"
        print("set fpsys at", p.fpsys)
        fpc=p.chip2fpc((id, chip))
    #fpc=  3096.602,  4194.244
        print("chip2fpc:\n\tid,chip=", id,chip,"\n\tfpc=",pixlib.set_decim_pts(fpc))
        msc=p.fpc2msc(fpc)
    #msc= (-10068.899462934856, 0.13953637467448429, 174.41686952444726)
        print("fpc2msc:\n\tfpc=", pixlib.set_decim_pts(fpc),"\n\tmsc=",pixlib.set_decim_pts(msc))
    
    
        p.aimpoint="AI1"
        print("set aimpoint at ",p.aimpoint)
        print("set tdet-sys at", p.tdetsys)
        id=3
        chip=783.8,894.0
        tdet=p.chip2tdet((id, chip))
        print("chip2tdet:\n\tid,chip=", id,chip,"\n\ttdet=",pixlib.set_decim_pts(tdet))
        if p.isokay != 1:
            print(p)

 

    # tdet= 1.0, 7168.0
            (id,chip)=p.tdet2chip(tdet)
            print("tdet2chip:\n\ttdet=",tdet,"\n\tid,chip=", id,chip)
            if p.isokay != 1:
                print(p)

 

            id=3
            chip_e=(783.8)
            fpc=p.chip2fpc((id,chip_e))
            print("chip2fpc:\n\tid,chip=", id,chip_e,"\n\tfpc=",fpc)
            print(" -> expected Error: need 'list' or 'tuple' type for 'chip' not the type of '<type 'float'>'.")
            if p.isokay != 1:
                print(p.errmsg+"\n")
        
            id_n_chip_e=3,(783.8,)
            fpc=p.chip2fpc(id_n_chip_e)
            print("chip2fpc:\n\tid,chip=", id_n_chip_e,"\n\tfpc=",fpc)     
            print(" -> expected Error: need 2-D for 'chip' but 1-D input.")
            if p.isokay != 1:
                print(p)


            fpc=p.chip2fpc((id,chip))
            print("chip2fpc:\n\tid,chip=", id,chip,"\n\tfpc=",fpc)     

            fpc_e=-35.7,981.5
            id_e_n_chip_e=p.fpc2chip(fpc_e)
            print("fpc2chip:\n\tfpc=", fpc_e,"\n\tid,chip=",id_e_n_chip_e)
            print(" -> expected Error: out of fpc bounds, ERROR: Can't find intersection on detector, Fail to do lsi ray to cpc transf.")
            if p.isokay != 1:
                print(p)
    
            id_e_n_chip_e=p.fpc2chip(fpc)
            print("fpc2chip:\n\tfpc=", fpc,"\n\tid,chip=", id_e_n_chip_e)
            print(" -> expected id,chip= (3, (783.78994148981633, 893.99118205816819))\n")
            if p.isokay != 1:
                print(p)
        
            gdp=p.chip2gdp((id,chip))
            print("chip2gdp:\n\tid,chip=", id,chip,"\n\tgdp=",gdp)
            if p.isokay != 1:
                print(p.errmsg+"\n")

            gdp=p.fpc2gdp(fpc)
            print("fpc2gdp:\n\tfpc=", fpc,"\n\tgdp=",gdp)
            if p.isokay != 1:
                print(p.errmsg+"\n")

            msc=p.fpc2msc(fpc)
            print("fpc2msc:\n\tfpc=", fpc,"\n\tmsc=",msc)
            fpc=p.msc2fpc(msc)
            print("msc2fpc:\n\tmsc=", msc,"\n\tfpc=",fpc)
            if p.isokay != 1:
                print(p.errmsg+"\n")


            aspoffs = 0.25,0.35
            sky=p.app_aspect(fpc, aspoffs)
            print("app_aspect:\n\tfpc=", fpc,"\n\tsky=",sky)
            print("\taspoffs=", aspoffs)
            print(" -> expected Error: need 3-D for 'asp-offsets' but 2-D input.")
            if p.isokay != 1:
                print(p.errmsg+"\n")

            aspoffs = 0.25,0.35,0.0
            sky=p.app_aspect(fpc, aspoffs)
            print("app_aspect:\n\tfpc=", fpc,"\n\tsky=",sky)
            print("\taspoffs=", aspoffs)
            if p.isokay != 1:
                print(p.errmsg+"\n")

            fpc=p.deapp_aspect(sky)
            print("deapp_aspect:\n\tsky=", sky,"\n\tfpc=",fpc)
            if p.isokay != 1:
                print(p.errmsg+"\n")


        p.close

    def test_extdchip(self):    
        pix = Pixlib("chandra", "geom.par")
        print("\n-- Test pixlib extended chip conversions ","-"*10)

        pix.detector = "ACIS"
        SIM_X   = -7.8090834371673E-01 
        SIM_Y   =  0.0000000000000E+00 
        SIM_Z   = -2.3358743446083E+02 
        DY_AVG  =  4.5452498327000E-02 
        DZ_AVG  =  2.1333081580000E-01 
        DTH_AVG = -2.1552640335000E-03 
 
        RA_NOM  =  2.4682474880853E+02 
        DEC_NOM = -2.4573378630717E+01 
            
        D_SIM = [ 0.0, -1*DY_AVG, -1*DZ_AVG ]
            
        pix.aimpoint = [ SIM_X, SIM_Y, SIM_Z ]
        pix.mirror = (  D_SIM, DTH_AVG )
        pix.fpsys  = "FP-1.1"
      
        """  
        expecting 
         ( 0, 1034.2003, 462.2634)
        """
        dxy = (3484.9970385478168, 4034.4782009383603)
        (id, chip) = pix.fpc2chip( dxy )
        print("\n1> fpc -> extended (id,chip) = (%2d," % id, ", ".join(["%.4f" % f for f in chip]), ')')

        fpc = pix.chip2fpc((id,chip))      
        if fpc[0] == None or  fpc[1] == None:
            print("1> extended (id,chip) -> fpc = (", ", ".join(["%10s" % f for f in fpc]), ')')
        else:
            print("1> extended (id,chip) -> fpc = (", ", ".join(["%10.4f" % f for f in fpc]), ')')


        """
        expecting
         ( 3, 1026.4890, 467.5626) 
        """
        dxy = (4624.37853503, 4045.39427581)     
        (id, chip) = pix.fpc2chip( dxy )
        print("\n2> fpc -> extended (id,chip) = (%2d," % id, ", ".join(["%.4f" % f for f in chip]), ')')
        fpc = pix.chip2fpc((id,chip))      
        if fpc[0] == None or  fpc[1] == None:
            print("2> extended (id,chip) -> fpc = (", ", ".join(["%10s" % f for f in fpc]), ')')
        else:
            print("2> extended (id,chip) -> fpc = (", ", ".join(["%10.4f" % f for f in fpc]), ')')


        pix.close
# ---------------
        pix = Pixlib("chandra", "geom.par")
        pix.detector = "ACIS"
        pix.fpsys  = "FP-3.1"
 

        """
        expecting
        fpc <- ext-chip (id= 2): (5153.070, 4046.554)
        """
        chip=(2, (-10.0,1.0) )
        fpc = pix.chip2fpc(chip)      
        if fpc[0] == None or  fpc[1] == None:
            print("\n3> extended (id,chip) -> fpc = (", ", ".join(["%10s" % f for f in fpc]), ')')
        else:
            print("\n3> extended (id,chip) -> fpc = (", ", ".join(["%10.4f" % f for f in fpc]), ')')
            (id, chip) = pix.fpc2chip(fpc)
            print("3> fpc -> extended (id,chip) = (%2d," % id, ", ".join(["%.4f" % f for f in chip]), ')')
  
        """
        expecting
        fpc <- ext-chip (id= 7): (3277.479, 5352.961)
        """
        chip=(7, (1054.5,1064.5) )
        fpc = pix.chip2fpc(chip)
        if fpc[0] == None or  fpc[1] == None:
            print("\n4> extended (id,chip) -> fpc = (", ", ".join(["%10s" % f for f in fpc]), ')')
        else:
            print("\n4> extended (id,chip) -> fpc = (", ", ".join(["%10.4f" % f for f in fpc]), ')')
        (id, chip) = pix.fpc2chipCoords( (7, fpc) )
        print("4> fpc (id=7)-> extended (chip) = (%2d," % id, ", ".join(["%.4f" % f for f in chip]), ')')
  


        """
        expecting
        fpc <-  chip (id= 4): (7388.901, 6400.933)
        """
        chip=(4, (0.5,0.5) )
        fpc = pix.chip2fpc(chip)
        print("\n5>  regular (id,chip) -> fpc = (", ", ".join(["%10.4f" % f for f in fpc]), ')')
        (id, chip) = pix.fpc2chip(fpc)
        print("5> fpc ->  regular (id,chip) = (%2d," % id, ", ".join(["%.4f" % f for f in chip]), ')')
  

 
        """
        expecting
        fpc -> chip (id= 2,    9.616,    1.111)
        """
        fpc=(5152.914, 4065.807)
        (id, chip) = pix.fpc2chip(fpc)
        print("\n6> fpc -> (id,chip) = (%2d," % id, ", ".join(["%.4f" % f for f in chip]), ')')
        fpc = pix.chip2fpc( (id,chip) )
        print("6>  regular (id,chip) -> fpc = (", ", ".join(["%10.4f" % f for f in fpc]), ')')

        pix.close

    def test_strings(self):    
         print("\n-- Test pixlib input string-types  ","-"*10)

    # expected okay
         pix = Pixlib("chandra", b"geom.par")  
         pix.close
         pix = Pixlib(u"chandra", b"geom.par")  
         pix.aimpoint=(-0.782,     0.000,  -237.500)
         print(pix.aimpoint) 
         if len(pix.errmsg):        # expected 'None'
             print(pix.errmsg)      # expected '', no error

         pix.aimpoint=b"ai1"
         if len(pix.errmsg):
             print(pix.errmsg)      # expected '', no error

    # expected errors
         pix.aimpoint="(-0.782,     0.000,  -237.500)" 
         print(pix.errmsg)      # expected error-message

         pix.aimpoint=u"anyaim"
         print(pix.errmsg)      # expected error-message

         pix.detector="anydet"
         print(pix.errmsg)      # expected error-message

         pix.close

    def test_failures(self):    
         print("\n-- Test pixlib expected failures  ","-"*10)
         allexp = 0
         count  = 0

    # expected okay
         pix = Pixlib(u"chandra", b"geom.par")  
 
      
         # test datatype, numeric types
         #  test SM55 bug report 
         pix.mirror=((10.0, -2.0, 1.0), np.float32(32)) 
         count += 1
         exp = "type np.float32 on stage-angle"
         if pix.isokay == 0:  # expect okay, should not get it here
             allexp += (exp == pix.errmsg)    # expected if strings equal
             print("%2i) '%s'\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
         else:
             exp = '((10.0, -2.0, 1.0), 32.0)'
             val = str(pix.mirror)
             print("%2i) '%s'\texpected: %s" % ( count, val, exp == val  ) )
             allexp += (exp == val)

         #  test np.int64()
         pix.mirror = ((10.0,-1.0,1.0), np.int64(20)) 
         count += 1
         exp = "type int64 on stage-angle"
         if pix.isokay == 0:  # expect okay, should not get it here
             allexp += (exp == pix.errmsg)    # expected if strings equal
             print("%2i) '%s'\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
         else:
             exp = '((10.0, -1.0, 1.0), 20.0)'
             val = str(pix.mirror)
             print("%2i) '%s'\texpected: %s" % ( count, val, exp == val  ) )
             allexp += (exp == val)

         #  test bool, np.int32(), int
         pix.mirror = ((bool(90), np.float32(-2), int(10) ), 20 ) 
         count += 1
         if pix.isokay == 0:  # expect okay, should not get it here
             allexp += (exp == pix.errmsg)    # expected if strings equal
             print("%2i) '%s'\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
         else:
             exp ='((1.0, -2.0, 10.0), 20.0)'
             val = str(pix.mirror)
             print("%2i) '%s'\texpected: %s" % ( count, val, exp == val  ) )
             allexp += (exp == val)
                
 
         #  test bool, np.int64(), bool
         pix.mirror = ((bool(0), np.int64(-2), bool(-10) ), 20 ) 
         count += 1
         if pix.isokay == 0:  # expect okay, should not get it here
             allexp += (exp == pix.errmsg)    # expected if strings equal
             print("%2i) '%s'\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
         else:
             exp ='((0.0, -2.0, 1.0), 20.0)'
             val = str(pix.mirror)
             print("%2i) '%s'\texpected: %s" % ( count, val, exp == val  ) )
             allexp += (exp == val)

         #  test bool
         pix.mirror = ((10.0,-1.0,1.0), bool(20)) 
         count += 1
         exp = "type bool on satge-angle"
         if pix.isokay == 0:  # expect okay, should not get it here
             allexp += (exp == pix.errmsg)    # expected if strings equal
             print("%2i) '%s'\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
         else:
             val = str(pix.mirror)
             exp = '((10.0, -1.0, 1.0), 1.0)'
             print("%2i) '%s'\texpected: %s" % ( count, val, exp == val  ) )
             allexp += (exp == val)


         # test datatype, string types
         #  test digit in string
         pix.mirror = ((0.0,-1.0,'1.0'),10.0)
         count += 1
         if pix.isokay == 0:  # not okay
            exp = "Error: invalid data type of 'stage-pos', '1.0', in (0.0, -1.0, '1.0')."
            allexp += (exp == pix.errmsg)    # expected if strings equal
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
 
         #  test string in string
         pix.mirror = ((0.0,-1.0,1.0), 'ImStrin')
         count += 1
         if pix.isokay == 0:
            exp = "Error: need non-string type for stage-angle in '((0.0, -1.0, 1.0), 'ImStrin')'."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )

     
        #  test numerical input on expected string
         pix.detector = 5
         count += 1
         exp = "Error: not strings  at '5'."
         if pix.isokay == 0:
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
                
         #  test numerical input on expected string
         pix.fpsys = 5.5
         count += 1
         if pix.isokay == 0:
            exp = "Error: not strings  at '5.5'."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
     
 
         # test dimension, expected 3-D 
         pix.mirror = ((0.0,-1.0, ), 10.0)
         count += 1
         if pix.isokay == 0:
            exp = "Error: need 3-D for 'stage-pos' but 2-D input."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
                                 
         # test input format, missing stg-angle
         pix.mirror = ((0.0,-1.0, 1.0), )
         count += 1
         if pix.isokay == 0:
            exp = "Error: expected input syntax: '((pos1,pos2,pos3), ang)' ."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
  
 
         fpc=pix.chip2fpc((2, ))
         count += 1
         if pix.isokay == 0:
            exp = "Error: expected input syntax: '(id, (chipx,chipy))' ."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )


         tdet=pix.chip2tdet((2,  None))
         count += 1
         if pix.isokay == 0:
            exp = "Error: need 'list' or 'tuple' type for 'chip' not the type of '<class 'NoneType'>'."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )

 
         pix.aimpoint=(-0.782,  0.000)
         count += 1
         if pix.isokay == 0:        # expected 'None'
            exp="Error: need 3-D for 'aimcoord' but 2-D input., at (-0.782, 0.0)'."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
      
                                  
            
         # test wrong format, expected list/tupe 
         pix.mirror = (0.0, 10.0)
         count += 1
         if pix.isokay == 0:
            exp="Error: need 'list' or 'tuple' type for 'stage-pos' not the type of '<class 'float'>'."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
 
         # test wrong input
         pix.grating="anyval"
         count += 1
         if pix.isokay == 0:        # expected 'None'
            exp="ERROR: Unsupported grating arm name, at 'anyval'."
            allexp += (exp == pix.errmsg) 
            print("%2i) '%s'\n\texpected: %s" % ( count, pix.errmsg, exp == pix.errmsg ) )
         
         pix.close
 
         if allexp==count:
             return 0     # return 0 if all expected TRUE  
         else:
             sys.exit(1)  # otherwise exit(1) 

if __name__ == "__main__":

    if "-u" in sys.argv[1:]:
        print(usage())        
    elif len(sys.argv)==1 or (len(sys.argv)==2 and ("-v" in sys.argv or "-q" in sys.argv ) ):
        unittest.main(defaultTest="TestPixlib")
    else:
        suite = unittest.TestSuite()
        for arg in sys.argv[1:]:
            if arg == "-aim":
                suite.addTest(TestPixlib("test_aimpoint")) 
                sys.argv.remove("-aim")
            elif arg == "-det":
                suite.addTest(TestPixlib("test_detector"))
                sys.argv.remove("-det")
            elif arg == "-def":
                suite.addTest(TestPixlib("test_default"))
                sys.argv.remove("-def")
            elif arg == "-gra":
                suite.addTest(TestPixlib("test_grating"))
                sys.argv.remove("-gra")
            elif arg == "-cor":
                suite.addTest(TestPixlib("test_transfer"))
                sys.argv.remove("-cor")
            elif arg == "-gen":
                suite.addTest(TestPixlib("test_general"))
                sys.argv.remove("-gen")
            elif arg == "-ext":
                suite.addTest(TestPixlib("test_extdchip"))
                sys.argv.remove("-ext")
            elif arg == "-str":
                suite.addTest(TestPixlib("test_strings"))
                sys.argv.remove("-str")
            elif arg == "-fai":
                suite.addTest(TestPixlib("test_failures"))
                sys.argv.remove("-fai")
            else:
                pass
      
        result = unittest.TextTestRunner(verbosity=2).run(suite)
  
        if result.wasSuccessful() :
            sys.exit(0)
        else:
            sys.exit(1)

