#!/usr/bin/perl
# 
#  Copyright (C) 2004-2005,2007,2019  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

@lib_paths = split /\s+/, $ENV{'ASCDS_INSTALL'};
$ASCDS_INSTALL = shift @lib_paths;
if ( ! -e 'test_utils.pl' )
{
   if ( -d "./datamodel" )
   {
      $TEXEC_DIR = "./datamodel";
   }
   elsif ( -d "$ASCDS_INSTALL/test/bin/datamodel" )
   {
      $TEXEC_DIR = "$ASCDS_INSTALL/test/bin/datamodel";
   }
}
unshift(@INC, $TEXEC_DIR);
require 'test_utils.pl';
$RM = find_inpath( "rm", $ENV{'PATH'}, ':' );
$MKDIR = find_inpath( "mkdir", $ENV{'PATH'}, ':' );
$PUNLEARN = find_inpath( 'punlearn', $ENV{'PATH'}, ':' );
$DMCOPY = find_inpath( "dmcopy", $ENV{'PATH'}, ':' );
$DMSTAT = find_inpath( "dmstat", $ENV{'PATH'}, ':' );
$VCOLS_TEST = find_inpath( 'vcols_test', ".:${TEXEC_DIR}", ':' );

use strict;
use vars '$RM', '$MKDIR', '$PUNLEARN', '$ASCDS_INSTALL', '$DMCOPY',  '$DMSTAT', 
         '$VCOLS_TEST';

##########################################################################
# PROGRAMMER - SET THESE VARIABLES!
my $id = '034';
my $semantic = 'VirtualCols';
my $script_name = 'dt${id}${semantic}.pl';

# Turn off buffering
select(STDOUT); $| = 1;
#
# Set some variables used by 'sort'
$ENV{'LC_COLLATE'}  = "en_US.ISO8859-1";
$ENV{'LC_MESSAGES'} = "C";

# Process Date
my $date = `date +%d%b%Y_%H:%M:%S`;
chop($date);
my ($junk,$start_hours,$start_min,$start_sec ) = $date =~ /(.*_)(\d*):(\d*):(\d*)/;

# set platform
my $platform = `uname`;
chop $platform;


##########################################################################
# Initialize some default values
my $USER    = $ENV{'USER'};
my $UPARM   = $ENV{'ASCDS_USER_PARAM'};
my $version = $ENV{'ASCDS_VERSION'};
my $home    = $ENV{'HOME'};
my $thisdir = $ENV{'PWD'};
my $testin  = "$ENV{'TESTIN'}/datamodel";
my $testout = "$ENV{'TESTOUT'}/datamodel";
my $testlog = "$ENV{'TESTLOG'}/datamodel";

my $testsav;
if ( -d "$ENV{'TESTSAV'}/datamodel" )
{
   $testsav = "$ENV{'TESTSAV'}/datamodel";
}
else
{
   $testsav = "$ASCDS_INSTALL/test/baseline/datamodel";
}
my $testrun = $ENV{'TESTRUN'};
my $printfile = *STDOUT;
my $caldb   = $ENV{'CALDB'};

#
#------------------------------------------------------------
# Load Arguments
my %inargs;
%inargs = &parse_args(\@ARGV);

#
# Process Arguments

# ***********
# -h = Help
# ***********
if ( $inargs{"-h"} )
{
   print "Usage: \n";
   print "   $script_name [ -iosfph ] \n";
   print "      -h   Help ( this text ) \n";
   print "\n";
   print "      -o   Specify test output directory \n";
   print "\n";
   print "      -i   Specify test input directory \n";
   print "\n";
   print "      -s   Specify test saved output directory \n";
   print "\n";
   print "      -f   Specify log file \n";
   print "\n";
   print "      -p   Specify printed output file \n";
   print "\n";
   print "      -k   Specify kernel, e.g., FITS or TEXT \n";
   print "\n\n";
   exit 0;
}

# ***********
# -i = TESTIN directory
# ***********
if ( $inargs{'-i'} )
{
   $testin = $inargs{'-i'};
}
# ***********
# -o = TESTOUT directory
# ***********
if ( $inargs{'-o'} )
{
   $testout = $inargs{'-o'};
}
# ***********
# -s = TESTSAV directory
# ***********
if ( $inargs{'-s'} )
{
   $testsav = $inargs{'-s'};
}
# make sure testing directories exist
unless ( -d $testlog ) { system( "$MKDIR -p $testlog" ); }
unless ( -d $testout ) { system( "$MKDIR -p $testout" ); }
unless ( -d $testin )
{
   print $printfile "ERROR: can't find input directory : $testin\n";
   exit 1;
}
unless ( -d $testsav )
{
   print $printfile "ERROR: can't find saved output directory : $testsav\n";
   exit 1;
}

# Clean up old log files
my $comstr = sprintf( "$RM -f $testlog/dt%.3d.log.*", $id );
system( $comstr );

# Open log files

# ***********
# -p = verbose output file
# ***********
if ( $inargs{'-p'} )
{
   unless ( open PRINT_OUTPUT, ">>$inargs{'-p'}" )
   {
      print STDERR "ERROR: Failed to open verbose output file : $inargs{'-p'}\n";
      exit 1;
   }
   $printfile = *PRINT_OUTPUT;
   # turn off buffering to print file
   select($printfile); $| = 1;
}

my $logfile;
my $logfname = "$testlog/dt${id}.log.$date";
# ***********
# -f = LOG file
# ***********
if ( $inargs{'-f'} )
{
   $logfname = $inargs{'-f'};
}
if ( open LOG_OUTPUT, ">>$logfname" )
{
   $logfile = *LOG_OUTPUT;
}
else
{
   print $printfile "ERROR: Failed to open log file : $logfname\n";
   exit 1;
}
# turn off buffering for the logfile
select($logfile); $| = 1;
select($printfile);

# get specified test kernel
my $testKernel = $ARGV[0];

# ***********
# -k = Kernel
# ***********
# anything to do here?
if ( $inargs{'-k'} )
{
   $testKernel = $inargs{'-k'};
}


# Parse the tolerance file
# This is only necessary if this test uses the dmlist_test or
# dmlist_header_test tests.
my $tolfile = sprintf("$testin/tol%.3d", $id);
my $tolkeysref = parse_tolfile( $tolfile );
my @tolkeys = @$tolkeysref;

######################### BEGIN TEST CODE ###################################

my $toolname = ""; # PROGRAMMER - put tool name here
my @testlist = 
( 
# PROGRAMMER - List the test ids here
  "1",
  "2",
  "3",
  "4",
  "5",
  "6",
  "7",
  "8",
  "9",
  "10",
  "11",
#  "12",
  "13",
  "14",
  "15",
  "16",
  "17",
  "18",
  "19",
  "20",
  "21",
  "22",
  "23",
  "24",
  "25",
  "26",
  "27",
  "28",
  "29",
  "30",
  "31",
  "32",
  "33",
  "34",
  "35",
  "36",
  "37",
  "38",
  "39",
  "40",
  "41",
  "42",
  "43",
  "44",
  "45"
);
my %test_blocks = 
( 
# PROGRAMMER - put lists of blocks and block types for each test
# eg:
# "1" => { "image" => [ 1, 3 ],
#          "table" => [ 2, 4 ] },
# "2" => { "list" => [ ] },...
 "1" => { "table" => [ 2 ], "list_header" => [ ] },
 "2" => { "table" => [ 2 ], "list_header" => [ ] },
 "3" => { "table" => [ 2 ], "list_header" => [ ] },
 "4" => { "table" => [ 2 ], "list_header" => [ ] },
 "5" => { "table" => [ 2 ], "list_header" => [ ] },
 "6" => { "table" => [ 2 ], "list_header" => [ ] },
 "7" => { "table" => [ 2 ], "list_header" => [ ] },
 "8" => { "table" => [ 2 ], "list_header" => [ ] },
 "9" => { "table" => [ 2 ], "list_header" => [ ] },
 "10" => { "table" => [ 2 ], "list_header" => [ ] },
 "11" => { "table" => [ 2 ], "list_header" => [ ] },
# "12" => { "table" => [ 2 ], "list_header" => [ ] },
 "13" => { "table" => [ 2 ], "list_header" => [ ] },
 "14" => { "table" => [ 2 ], "list_header" => [ ] },
 "15" => { "image" => [ 1 ], "list_header" => [ ] },
 "16" => { "image" => [ 1 ], "list_header" => [ ] },
 "17" => { "image" => [ 1 ], "list_header" => [ ] },
 "18" => { "image" => [ 1 ], "list_header" => [ ] },
 "19" => { "image" => [ 1 ], "list_header" => [ ] },
 "20" => { "image" => [ 1 ], "list_header" => [ ] },
 "21" => { "image" => [ 1 ], "list_header" => [ ] },
 "22" => { "image" => [ 1 ], "list_header" => [ ] },
 "23" => { "image" => [ 1 ], "list_header" => [ ] },
 "24" => { "image" => [ 1 ], "list_header" => [ ] },
 "25" => { "image" => [ 1 ], "list_header" => [ ] },
 "26" => { "table" => [ 2 ], "list_header" => [ ] },
 "27" => { "image" => [ 1 ], "list_header" => [ ] },
 "28" => { "table" => [ 2 ], "list_header" => [ ] },
 "29" => { "table" => [ 2 ], "list_header" => [ ] },
 "30" => { "table" => [ 2 ], "list_header" => [ ] },
 "31" => { "table" => [ 2 ], "list_header" => [ ] },
 "32" => { "table" => [ 2 ], "list_header" => [ ] },
 "33" => { "table" => [ 2 ], "list_header" => [ ] },
 "34" => { "list" => [ ] },
 "35" => { "list" => [ ] },
 "36" => { "list" => [ ] },
 "37" => { "list" => [ ] },
 "38" => { "list" => [ ] },
 "39" => { "list" => [ ] },
 "40" => { "list" => [ ] },
 "41" => { "text" => [ ] },
 "42" => { "text" => [ ] },
 "43" => { "text" => [ ] },
 "44" => { "text" => [ ] },
 "45" => { "image" => [ 1 ] }
);
my %test_words = 
(
# PROGRAMMER - put list of words to replace in a text output file
#              This is only necessary if using the diff test for text files
# eg:
# $version => "ASCDS_VERSION"
 
);

		     
#my $toolpath = find_inpath($toolname, $ENV{'PATH'}, ':');
#unless ( $toolpath )
#{
#   print $printfile "ERROR: Unable to find tool : $toolname\n";
#   exit 1;
#}

# Set parameters to defaults
system( "$PUNLEARN $toolname" );


dt_print_mt_header( $id, $semantic, $logfile, $printfile );

# Prepare for main loop
my @faillist = ( );
my $time_user_total = 0;
my $time_sys_total = 0;
my $rc;
my $err_msg = "";

my %exten = ("FITS" => "fits",
             "TEXT" => "dtf" );
my %copyOpts = ("TEXT" => "text/dtf");
my %links = ("1" => "indt034m001",
             "11" => "indt034m011",
             "30" => "indt034m030",
             "42" => "indt034m042" ); 
my $kernelName = "";
my $line = "";
my $pid;
my @kernels = "";
my $test_string;
my $in_ext = "fits";
my $out_ext = "fits";
my $source_ext = "fits";
#########################################################################
# start the child process to get the kernel list
$pid = open( INHANDLE, "getkernellist |");

# get the output of the child process
$line = <INHANDLE>;

# break the output line into individual datum and put
# into list of kernel names
@kernels = split(/\s+/, $line);

close(INHANDLE);

# Control looping over each installed kernel
# by checking for kernel argument.
# If present, perform tests only the specified kernel
# If absent, perform tests on each installed kernel

foreach $kernelName (@kernels)
{
  my $errFlag = 1;

  if (($testKernel ne "") && ($kernelName ne $testKernel))
  {
    next;
  }

  # Loop over all tests
  foreach my $testid (@testlist)
  {
    # Init per-test error flag
    my $mismatch = 0;
    $err_msg = "";
    my $time_user;
    my $time_sys;
    my $skipDelete = 0;
    my $linkname = "";

    # change extension based on kernel
    $in_ext = $exten{$kernelName};

    # delete old outputs
    my $basename = sprintf "dt%.3dm%.3d", $id, $testid;
    system("$RM -f $testout/${basename}*");

    my $block_typesref = $test_blocks{$testid};
    my %block_types = %$block_typesref;

    my $infile = "$testin/in${basename}.$in_ext";
    my $inbase = sprintf "$testin/indt%.3d", $id;
    my $tolerance = sprintf "$testin/tolerance%.3d", $id;

    # if testing any kernel other than FITS, the FITS input files need to
    # be copied into the present kernel's format
    if ( $kernelName ne "FITS" )
    {
      my $sourceFile = "$testin/in${basename}.$source_ext";
      my $copyFile   = "$testout/in${basename}.$in_ext";
      my $options    = "opt=all";
      my $kernelOpts = $copyOpts{"$kernelName"};

      if ( ! -e "$infile" )
      {
        my $linkbase = "";
        my $teststring = "$DMCOPY $sourceFile $copyFile\"[opt kernel=$kernelOpts]\" $options";

        # does a previous copy file exist?
        if ($testid == 11 || $testid == 25 )
        {
          $linkbase = $links{"11"};
          $linkname = "$testout/${linkbase}.$in_ext";
        }
        elsif ($testid >= 30 && $testid <= 40 )
        {
          $linkbase = $links{"30"};
          $linkname = "$testout/${linkbase}.$in_ext";
        }
        elsif ($testid >= 42 && $testid <= 44 )
        {
          $linkbase = $links{"42"};
          $linkname = "$testout/${linkbase}.$in_ext";
        }
        elsif ($testid == 45 )
        {
          $linkbase = $links{"45"};
          $linkname = "$testout/${linkbase}.$in_ext";
        }
	else
        {
          $linkbase = $links{"1"};
          $linkname = "$testout/${linkbase}.$in_ext";
        }

        # if not, make one
        if ( ! -e $copyFile )
        {
          if ( $linkname ne "" )
          {
            if ( -e $linkname )
            {
              # create link
              $teststring = "ln -s $linkname $copyFile";
            }
          }

          # copy fits file to text/dtf file or create link
          ($rc, $time_user, $time_sys) = system($teststring);
        }

        $infile = $copyFile;

        # Tests 1-10, 12-24, 26-29, and 41 use the same input file.
        # Tests 11 and 25 use the same input file
        # Tests 30 - 40 use the same input file
	# Tests 42 - 44 use the same input file
	# Don't delete generated input file
        if ( $testid == 1 || $testid == 11 || $testid == 30 || $testid == 42 )
        {
          $skipDelete = 1;
        }
      }
      else
      {
        $skipDelete = 1;
      }
    }

    # PROGRAMMER - put output filename here
    my $outfile = "$testout/${basename}.fits";
    # PROGRAMMER - put saved output filename here
    my $savfile = "$testsav/${basename}.fits.gz";
    # PROGRAMMER - put name of diff file here, if needed
    my $test_diff = "$testout/${basename}.diff";
    # PROGRAMMER - put test string here
    my $test_string = "";

    if ( $testid == 1 )
    {
       $test_string = "$DMCOPY $infile\"[ra=233.5:233.7,dec=23.4:23.6]\" $outfile";
    }
    elsif ( $testid == 2 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.5:233.7,dec=23.4:23.6][cols ra,dec]\" $outfile";
    }
    elsif ( $testid == 3 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.5:233.7,dec=23.4:23.6][cols ra,dec,*]\" $outfile";
    }
    elsif ( $testid == 4 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.5:233.7,dec=23.4:23.6][cols eqpos]\" $outfile";
    }
    elsif ( $testid == 5 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.5:233.7,dec=23.4:23.6][cols sky]\" $outfile";
    }
    elsif ( $testid == 6 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.5:233.7][cols sky]\" $outfile";
    }
    elsif ( $testid == 7 )
    {
      $test_string = "$DMCOPY $infile\"[dec=23.4:23.6][cols sky]\" $outfile";
    }
    elsif ( $testid == 8 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.5:233.7][cols dec]\" $outfile";
    }
    elsif ( $testid == 9 )
    {
      $test_string = "$DMCOPY $infile\"[dec=23.4:23.6][cols ra]\" $outfile";
    }
    elsif ( $testid == 10 )
    {
      $test_string = "$DMCOPY $infile\"[ra>233.5,dec<23.4][cols sky]\" $outfile";
    }
    elsif ( $testid == 11 )
    {
      $test_string = "$DMCOPY $infile\"[ra>233.5,dec<23.4][cols sky]\" $outfile";
    }
    # This one doesn't work -- should it?
    #elsif ( $testid == 12 )
    #{
    #   $test_string = "$DMCOPY infile\"[ra=233.5:233.7,dec=23.4:23.6][cols c1=ra,c2=y]\" $outfile";
    #}
    elsif ( $testid == 13 )
    {
      $test_string = "$DMCOPY $infile\"[exclude ra>233.5][cols sky]\" $outfile";
    }
    elsif ( $testid == 14 )
    {
      $test_string = "$DMCOPY $infile\"[#row=1:100,dec=23.4:23.6][cols ra]\" $outfile";
    }
    elsif ( $testid == 15 )
    {
      $test_string = "$DMCOPY $infile\"[bin ra=233.42:233.99:0.01,dec=23.23:23.78:0.01]\" $outfile";
    }
    elsif ( $testid == 16 )
    {
      $test_string = "$DMCOPY $infile\"[bin ra=233.42:233.99:#200,dec=23.23:23.78:#200]\" $outfile";
    }
    elsif ( $testid == 17 )
    {
      $test_string = "$DMCOPY $infile\"[bin ra=0.01,dec=0.01]\" $outfile";
    }
    elsif ( $testid == 18 )
    {
      $test_string = "$DMCOPY $infile\"[bin ra=::0.01,dec=::0.01]\" $outfile";
    }
    elsif ( $testid == 19 )
    {
      $test_string = "$DMCOPY $infile\"[bin ra=233.6::0.01,dec=:23.5:0.01]\" $outfile";
    }
    elsif ( $testid == 20 )
    {
      $test_string = "$DMCOPY $infile\"[bin ra,dec][opt type=i4]\" $outfile";
    }
    elsif ( $testid == 21 )
    {
      $test_string = "$DMCOPY $infile\"[bin eqpos][opt type=i4]\" $outfile";
    }
    elsif ( $testid == 22 )
    {
      $test_string = "$DMCOPY $infile\"[bin eqpos=0.01]\" $outfile";
    }
    elsif ( $testid == 23 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.7:233.8][bin ra=::0.01,dec=::0.01]\" $outfile";
    }
    elsif ( $testid == 24 )
    {
      $test_string = "$DMCOPY $infile\"[ra=233.7:][bin eqpos=0.01]\" $outfile";
    }
    elsif ( $testid == 25 )
    {
      $test_string = "$DMCOPY $infile\"[bin ra=::#200,dec=::#200]\" $outfile";
    }
    elsif ( $testid == 26 )
    {
      $test_string = "$DMCOPY $infile\"[cpcx=10.5:15.72,cpcy=7.894:13.21][cols cpc]\" $outfile";
    }
    elsif ( $testid == 27 )
    {
      $test_string = "$DMCOPY $infile\"[bin cpc=0.1]\" $outfile";
    }
    elsif ( $testid == 28 )
    {
      $test_string = "$DMCOPY $infile\"[cpcy>13.5,dec=23.4:23.6]\" $outfile";
    }
    elsif ( $testid == 29 )
    {
      $test_string = "$DMCOPY $infile\"[cpcy>13.5,dec=23.4:23.6][cols dec,cpcy]\" $outfile";
    }
    elsif ( $testid == 30 )
    {
      $test_string = "$DMCOPY $infile\"[energy=1500:7300]\" $outfile";
    }
    elsif ( $testid == 31 )
    {
      $test_string = "$DMCOPY $infile\"[energy=1500:7300][cols energy]\" $outfile";
    }
    elsif ( $testid == 32 )
    {
      $test_string = "$DMCOPY $infile\"[energy=1500:7300][cols energy,pi]\" $outfile";
    }
    elsif ( $testid == 33 )
    {
      $test_string = "$DMCOPY $infile\"[energy=1500:7300,9200:12000][cols energy]\" $outfile";
    }
    elsif ( $testid == 34 )
    {
      $test_string = "$DMCOPY $infile\"[bin energy=1500:7300:10]\" $outfile";
    }
    elsif ( $testid == 35 )
    {
      $test_string = "$DMCOPY $infile\"[bin energy=1500:7300:#100]\" $outfile";
    }
    elsif ( $testid == 36 )
    {
      $test_string = "$DMCOPY $infile\"[bin energy=10]\" $outfile";
    }
    elsif ( $testid == 37 )
    {
      $test_string = "$DMCOPY $infile\"[bin energy=:7300:10]\" $outfile";
    }
    elsif ( $testid == 38 )
    {
      $test_string = "$DMCOPY $infile\"[bin energy=::10]\" $outfile";
    }
    elsif ( $testid == 39 )
    {
      $test_string = "$DMCOPY $infile\"[bin energy=1500:2500]\" $outfile";
    }
    elsif ( $testid == 40 )
    {
      $test_string = "$DMCOPY $infile\"[energy=1500:2500][bin energy]\" $outfile";
    }
    elsif ( $testid == 41 )
    {
      $outfile = "$testout/${basename}.txt";
      $savfile = "$testsav/${basename}.txt";
      $test_string = "( $VCOLS_TEST $infile | tee $outfile )";
    }
    elsif ( $testid == 42 )
    {
      # Bug# SL159:  physical column filter producing (0,0) rows in virtual column
	$outfile = "$testout/${basename}.txt";
	$savfile = "$testsav/${basename}.txt";
	$test_string = "( $DMSTAT $infile\"[sky=circle(3888.5,4198.5,3)][cols eqpos]\" sigma- | tee $outfile )";
    }
    elsif ( $testid == 43 )
    {
      # Bug# SL159:  physical column filter producing (0,0) rows in virtual column
	$outfile = "$testout/${basename}.txt";
	$savfile = "$testsav/${basename}.txt";
	$test_string = "( $DMSTAT $infile\"[sky=annulus(3888.5,4198.5,2.9,3)][cols eqpos]\" sigma- | tee $outfile )";
    }
     elsif ( $testid == 44 )
    {
      # Bug# SL159:  physical column filter producing (0,0) rows in virtual column
	$outfile = "$testout/${basename}.txt";
	$savfile = "$testsav/${basename}.txt";
	$test_string = "( $DMSTAT $infile\"[energy=500:7000,sky=annulus(3888.5,4198.5,2.9,3)][cols eqpos]\" sigma- | tee $outfile )";
    }
     elsif ( $testid == 45 )
    {
      # Bug# SL185, because of SL159 :  filter&bin bug
	$test_string = "( $DMCOPY $infile\"[x=3407.5:4316.5, y=3537.5:4284.5][bin sky=1]\" $outfile )";
    }

   # run the tool
    $test_string .= " >> $logfname 2>&1";
    dt_print_test_header( $testid, $test_string, $logfile, $printfile, $kernelName );
    close $logfile;
    ($rc, $time_user, $time_sys) = dt_run_test($test_string);

    open $logfile, ">>$logfname";
    select($logfile); $| = 1; select($printfile);
    unless ( $rc == 0 )
    {
      $err_msg = "ERROR: $toolname failed to run\n";
    }
   
    if ( $block_types{"table"} )
    {
      my @blocks = @{$block_types{"table"}};
      foreach my $block ( @blocks )
      {
        $err_msg .= dmdiff_test( $outfile, $savfile, $block, "$tolerance", $logfname, "${test_diff}_table_${block}" );
      }
    }
   if ( $block_types{"image"} )
   {
      my @blocks = @{$block_types{"image"}};
      foreach my $block ( @blocks )
      {
        $err_msg .= dmimgcalc_test( $outfile, $savfile, $block, "$tolerance", $logfname, "${test_diff}_image_${block}" );
      }
    }
    if ( $block_types{"list"} )
    {
      $err_msg .= dmlist_test( $outfile, "${outfile}_filt_list", $savfile, "$testout/${basename}_savfilt_list", \@tolkeys, $logfname, "${test_diff}_list", "" );
    }
    if ( $block_types{"list_header"} )
    {
      $err_msg .= dmlist_header_test( $outfile, "${outfile}_filt_list_header", $savfile, "$testout/${basename}_savfilt_list_header", \@tolkeys, $logfname, "${test_diff}_list_header" );
    }
    if ( $block_types{"text"} )
    {
      $err_msg .= diff_test( $outfile, "${outfile}_filt_text", $savfile, "$testout/${basename}_savfilt_text", \%test_words, $logfname, "${test_diff}_text" );
    }

    unless ( $err_msg eq "" )
    {
      # Put the kernel name in front of the list of failed tests
      if ( $errFlag == 1 )
      {
        push (@faillist, $kernelName);
        $errFlag = 0;
      }
      push( @faillist, $testid );
      $mismatch = 1;
    }

   $time_user_total += $time_user;
   $time_sys_total  += $time_sys;
   dt_print_test_footer( $testid, $mismatch, $err_msg, $time_user, $time_sys, $logfile, $printfile );

    # delete any input files created just for this kernel
    if ( $kernelName ne "FITS" )
    {
      if ( $skipDelete == 0 )
      {
        system("$RM -f $infile");
      }
      if ($testid == 25 || $testid == 40 || $testid == 41)
      {
          system("$RM -f $linkname");
      }
    }
  }
}

if ( scalar @faillist > 0 )
{
   $rc = 1;
}
else
{
   $rc = 0;
}
$err_msg = "@faillist";
dt_print_mt_footer( $id, $semantic, $rc, $err_msg, $time_user_total, $time_sys_total, $logfile, $printfile );

close $logfile;
close $printfile;
exit $rc;
